/**
 * Copy(?)Right! Pablo Virolainen (pablo@ipi.fi) 2008
 * 
 * See COPYING.LGPL (LGPL)
 * This is part of ginotify (0.1)
 */

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <glib-object.h>
#include <glib.h>
#include "ginotify.h"

struct _gnotify_entry {
	GINOTIFYFunc	function;
	gpointer	data;
};


static
gint funcs_comparator(gconstpointer a,
		      gconstpointer b,
		      gpointer user_data)
{
	if (a < b)
	{
		return -1;
	}

	if (a == b)
  	{
		return 0;
	}
	return 1;
}

static
gint funcs_find_comparator(gconstpointer _a,
			   gconstpointer _b)
{
	return funcs_comparator(_b,_a,NULL);
}


static
gboolean notify_incoming(GIOChannel *source,
			 GIOCondition condition,
			 gpointer data)
{
	int red;
	GINotify *notify=data;
	char buf[65536];
	struct inotify_event *event;


	while ((red=read(notify -> fd, buf, sizeof (buf)))>0) {
		event = (struct inotify_event*) buf;
		while (red>0 && event -> len > 0) {
		  struct _gnotify_entry *entry = 
			g_tree_search(notify -> funcs,
				      funcs_find_comparator,
				      (void*)(event -> wd));
			
			if (entry != NULL && entry -> function != NULL) {
				if (! entry -> function (notify,
							 event -> wd,
							 event -> mask,
							 event -> cookie,
							 event -> name,
							 entry -> data)) {
					notify_remove (notify,
						       event -> wd);
				}
			}
			red-=event -> len;
			event = ((void*) event) + event->len;
		}
	}
	return TRUE;
}


static
void funcs_value_destroy_func (gpointer data)
{
	g_free(data);
}

GINotify* notify_new() {
	int fd = inotify_init ();
	int read_size=-1;
	GINotify* notify = g_new0(GINotify,1);

	g_return_val_if_fail (notify != NULL, NULL);

	GIOChannel* channel = g_io_channel_unix_new(fd);

	notify -> channel = channel;
	notify -> fd = fd;
	notify -> read_size = read_size;
	notify -> funcs = g_tree_new_full(funcs_comparator,
					  NULL,
					  NULL,
					  funcs_value_destroy_func);
	
	g_io_add_watch(channel,G_IO_IN, notify_incoming, notify);

	return notify;
}

void notify_free(GINotify* notify) {
	g_io_channel_shutdown(notify -> channel, TRUE, NULL);
	g_tree_destroy(notify -> funcs);
	g_free(notify);
}

gint notify_add(GINotify	*notify,
		const char	*path,
		guint		mask,
		GINOTIFYFunc	function,
		gpointer	data)
{
	gint id;
	struct _gnotify_entry *new_entry; 
	g_return_val_if_fail (notify != NULL,-1);

	new_entry = g_new0(struct _gnotify_entry,1);

	g_return_val_if_fail (notify != NULL,-1);
	
	id = inotify_add_watch(notify->fd, path, mask);
	
	new_entry -> function = function;
	new_entry -> data = data;
	g_tree_insert(notify->funcs, (void*)id, new_entry);

	return id;
}

gboolean notify_remove(GINotify	*notify,
		       gint id)
{
	g_return_val_if_fail (notify != NULL,FALSE);
	
	g_tree_remove( notify -> funcs, (void*) id);
	
	return inotify_rm_watch(notify -> fd,id);	
}
